@@ -105,16 +105,16 @@ testProxy4 = proxyFun4
 
 testProxy4Errors =
 #if __GLASGOW_HASKELL__ >= 914
-  ["Expected: Proxy ((2 * 0) + 4)"
+  ["Expected: Proxy ((2 * y0) + 4)"
   ,"  Actual: Proxy 2"
   ]
 #elif __GLASGOW_HASKELL__ >= 900
   ["Expected: Proxy 2 -> ()"
-  ,"  Actual: Proxy ((2 * 0) + 4) -> ()"
+  ,"  Actual: Proxy ((2 * y0) + 4) -> ()"
   ]
 #else
   ["Expected type: Proxy 2 -> ()"
-  ,"Actual type: Proxy ((2 * 0) + 4) -> ()"
+  ,"Actual type: Proxy ((2 * y0) + 4) -> ()"
   ]
 #endif
 
@@ -123,16 +123,16 @@ testProxy5 = proxyFun4
 
 testProxy5Errors =
 #if __GLASGOW_HASKELL__ >= 914
-  ["Expected: Proxy ((2 * y0) + 4)"
+  ["Expected: Proxy ((2 * y1) + 4)"
   ,"  Actual: Proxy 7"
   ]
 #elif __GLASGOW_HASKELL__ >= 900
   ["Expected: Proxy 7 -> ()"
-  ,"  Actual: Proxy ((2 * y0) + 4) -> ()"
+  ,"  Actual: Proxy ((2 * y1) + 4) -> ()"
   ]
 #else
   ["Expected type: Proxy 7 -> ()"
-  ,"Actual type: Proxy ((2 * y0) + 4) -> ()"
+  ,"Actual type: Proxy ((2 * y1) + 4) -> ()"
   ]
 #endif
 
@@ -592,3 +592,21 @@ t99_errors =
      )
   ]
 #endif
+
+t113 :: Proxy a -> Proxy b -> Proxy ((2 * a) + b) -> Proxy 5
+t113 _ _ = id
+
+t113_errors =
+#if __GLASGOW_HASKELL__ >= 914
+  ["Expected: Proxy 5"
+  ,"  Actual: Proxy ((2 * a) + b)"
+  ]
+#elif __GLASGOW_HASKELL__ >= 900
+  ["Expected: Proxy ((2 * a) + b) -> Proxy 5"
+  ,"  Actual: Proxy 5 -> Proxy 5"
+  ]
+#else
+  ["Expected type: Proxy ((2 * a) + b) -> Proxy 5"
+  ,"Actual type: Proxy 5 -> Proxy 5"
+  ]
+#endif
